package aspectjtest;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/** 
 * a generic aspect, we've used descriptive role names for the type variables
 * (Parent and Child) but you could use anything of course
 */
public abstract aspect ParentChildRelationship<Parent,Child> {

  /** generic interface implemented by parents */   
  interface ParentHasChildren<C extends ChildHasParent>{
    List<C> getChildren();
    void addChild(C child);
    void removeChild(C child);
  }

  /** generic interface implemented by children */
  interface ChildHasParent<P extends ParentHasChildren>{
    P getParent();
    void setParent(P parent);
  }

  /** ensure the parent type implements ParentHasChildren<child type> */
  declare parents: Parent implements ParentHasChildren<Child>;

  /** ensure the child type implements ChildHasParent<parent type> */
  declare parents: Child implements ChildHasParent<Parent>;

  // Inter-type declarations made on the *generic* interface types to provide 
  // default implementations.

  /** list of children maintained by parent */
  private List<C> ParentHasChildren<C>.children = new ArrayList<C>();

  /** reference to parent maintained by child */        
  private P ChildHasParent<P>.parent;

  /** Default implementation of getChildren for the generic type ParentHasChildren */
  public List<C> ParentHasChildren<C>.getChildren() {
        return Collections.unmodifiableList(children);  
  }

  /** Default implementation of getParent for the generic type ChildHasParent */    
  public P ChildHasParent<P>.getParent() {
       return parent;
  }

  /** 
    * Default implementation of addChild, ensures that parent of child is
    * also updated.
    */ 
  public void ParentHasChildren<C>.addChild(C child) {
       if (child.parent != null) {
         child.parent.removeChild(child);
       }
       children.add(child);
       child.parent = this;
    }

   /**
     * Default implementation of removeChild, ensures that parent of
     * child is also updated.
     */
   public void ParentHasChildren<C>.removeChild(C child) {
       if (children.remove(child)) {
         child.parent = null;
       }
    }

    /**
      * Default implementation of setParent for the generic type ChildHasParent.
      * Ensures that this child is added to the children of the parent too.
      */
    public void ChildHasParent<P>.setParent(P parent) {
       parent.addChild(this);
    }

    /**
      * Matches at an addChild join point for the parent type P and child type C
      */    
    public pointcut addingChild(Parent p, Child c) :
      execution(* ParentHasChildren.addChild(ChildHasParent)) && this(p) && args(c);
      
    /**
      * Matches at a removeChild join point for the parent type P and child type C
      */    
    public pointcut removingChild(Parent p, Child c) :
      execution(* ParentHasChildren.removeChild(ChildHasParent)) && this(p) && args(c);

}